from fa_purity.frozen import (
    freeze,
)
from fa_purity.json_2.primitive import (
    JsonPrimitiveFactory,
    Primitive,
)
from fa_purity.json_2.value import (
    JsonValue,
    JsonValueFactory,
    Unfolder,
)


def _prim_value(value: Primitive) -> JsonValue:
    return JsonValue.from_primitive(JsonPrimitiveFactory.from_raw(value))


def test_from_any() -> None:
    json_obj = freeze(
        {
            "foo": JsonValue.from_json(
                freeze(
                    {
                        "nested": JsonValue.from_list(
                            tuple([_prim_value("hi"), _prim_value(99)])
                        ),
                    }
                )
            )
        }
    )
    json_obj_from_raw = (
        JsonValueFactory.from_any({"foo": {"nested": ["hi", 99]}})
        .bind(Unfolder.to_json)
        .unwrap()
    )
    assert json_obj == json_obj_from_raw



from __future__ import (
    annotations,
)

from dataclasses import (
    dataclass,
)
from fa_purity.result import (
    Result,
)
from typing import (
    Callable,
    Generic,
    NoReturn,
    Optional,
    Type,
    TypeVar,
    Union,
)

_A = TypeVar("_A")
_B = TypeVar("_B")


@dataclass(frozen=True)
class Maybe(Generic[_A]):
    _value: Result[_A, None]

    @staticmethod
    def from_value(value: _A) -> Maybe[_A]:
        return Maybe(Result.success(value))

    @staticmethod
    def from_optional(value: Optional[_A]) -> Maybe[_A]:
        if value is None:
            return Maybe(Result.failure(value))
        return Maybe(Result.success(value))

    @staticmethod
    def from_result(result: Result[_A, None]) -> Maybe[_A]:
        return Maybe(result)

    @staticmethod
    def empty(_type: Optional[Type[_A]] = None) -> Maybe[_A]:
        return Maybe.from_optional(None)

    def to_result(self) -> Result[_A, None]:
        return self._value

    def map(self, function: Callable[[_A], _B]) -> Maybe[_B]:
        return Maybe(self._value.map(function))

    def bind(self, function: Callable[[_A], Maybe[_B]]) -> Maybe[_B]:
        return Maybe(self._value.bind(lambda a: function(a).to_result()))

    def bind_optional(
        self, function: Callable[[_A], Optional[_B]]
    ) -> Maybe[_B]:
        return self.bind(lambda a: Maybe.from_optional(function(a)))

    def lash(self, function: Callable[[], Maybe[_A]]) -> Maybe[_A]:
        return Maybe(self._value.lash(lambda _: function().to_result()))

    def unwrap(self) -> Union[_A, NoReturn]:
        return self._value.unwrap()

    def failure(self) -> Union[None, NoReturn]:
        return self._value.unwrap_fail()

    def value_or(self, default: _B) -> Union[_A, _B]:
        return self._value.value_or(default)

    def or_else_call(self, function: Callable[[], _B]) -> Union[_A, _B]:
        return self._value.or_else_call(function)



from fa_purity.cmd import (
    Cmd,
)
from fa_purity.frozen import (
    FrozenDict,
    FrozenList,
)
from fa_purity.json import (
    JsonObj,
    JsonValue,
    UnfoldedJVal,
)
from fa_purity.maybe import (
    Maybe,
)
from fa_purity.pure_iter import (
    PureIter,
)
from fa_purity.result import (
    Result,
    ResultE,
)
from fa_purity.stream import (
    Stream,
)

__version__ = "1.33.1"
__all__ = [
    "PureIter",
    "Stream",
    "JsonObj",
    "JsonValue",
    "UnfoldedJVal",
    "Cmd",
    "FrozenDict",
    "FrozenList",
    "Result",
    "ResultE",
    "Maybe",
]



from __future__ import (
    annotations,
)

from dataclasses import (
    dataclass,
)
from deprecated import (
    deprecated,
)
from fa_purity._patch import (
    Patch,
)
import sys
from typing import (
    Callable,
    Generic,
    NoReturn,
    TypeVar,
)

_A = TypeVar("_A")
_B = TypeVar("_B")


@dataclass(frozen=True)
class Cmd(Generic[_A]):
    # Equivalent to haskell IO type
    # unsafe_unwrap not included as method for discouraging its use
    _value: Patch[Callable[[], _A]]

    @staticmethod
    def from_cmd(value: Callable[[], _A]) -> Cmd[_A]:
        return Cmd(Patch(value))

    @staticmethod
    def new_cmd(action: Callable[[CmdUnwrapper], _A]) -> Cmd[_A]:
        return Cmd(Patch(lambda: action(_unwrapper)))

    def map(self, function: Callable[[_A], _B]) -> Cmd[_B]:
        return Cmd(Patch(lambda: function(self._value.unwrap())))

    def bind(self, function: Callable[[_A], Cmd[_B]]) -> Cmd[_B]:
        return Cmd(
            Patch(lambda: function(self._value.unwrap())._value.unwrap())
        )

    def apply(self, wrapped: Cmd[Callable[[_A], _B]]) -> Cmd[_B]:
        return wrapped.bind(lambda f: self.map(f))

    def compute(self) -> NoReturn:
        self._value.unwrap()
        sys.exit(0)

    def __add__(self, other: Cmd[_B]) -> Cmd[_B]:
        return self.bind(lambda _: other)


@dataclass(frozen=True)
class _CmdUnwrapper:
    pass


@dataclass(frozen=True)
class CmdUnwrapper:
    """
    Object that brings a method that allows cmd execution,
    equivalent to `unsafe_unwrap`.

    Instances can not be created by the user by design.
    Only through the `Cmd.new_cmd` the user can use an
    instance of this type.
    """

    # Do not build any public constructors or instances
    # This obj is only accessible in the action context through the `new_cmd` builder
    _inner: _CmdUnwrapper

    def act(self, action: Cmd[_A]) -> _A:
        """
        [WARNING] Not a pure function.

        This method is more safe than `unsafe_unwrap` since it
        wraps the result into another `Cmd`, but is possible to
        use it incorrectly in places where a pure function is
        expected e.g. `PureIter.map` and cause unexpected bugs.
        """
        return action._value.unwrap()

    @staticmethod
    @deprecated("DO NOT CALL: instead use the `act` method of a `CmdUnwrapper` instance")  # type: ignore[misc]
    def unwrap(action: Cmd[_A]) -> _A:
        return action._value.unwrap()


_unwrapper = CmdUnwrapper(_CmdUnwrapper())


@deprecated("NEW API: use `Cmd.new_cmd` instead")  # type: ignore[misc]
def new_cmd(action: Callable[[CmdUnwrapper], _A]) -> Cmd[_A]:
    return Cmd(Patch(lambda: action(_unwrapper)))


def unsafe_unwrap(action: Cmd[_A]) -> _A:
    # This is an unsafe constructor (type-check cannot ensure its proper use)
    # Do not use until is strictly necessary
    # WARNING: this is equivalent to compute, and will execute the Cmd
    #
    # Only use when all actions (Cmd[_A]) result in the same output instance (_A)
    # and side effects are not present or negligible.
    # e.g. unwrap a cmd when used on a cached function definition
    #
    # [NOTICE] Do not use this function for defining a new Cmd, use `new_cmd` instead
    return action._value.unwrap()



from fa_purity.union import (
    Coproduct,
    CoproductFactory,
    CoproductTransform,
)


def test_equality() -> None:
    x: Coproduct[int, str] = Coproduct.inl(23)
    y: Coproduct[str, int] = Coproduct.inr(23)
    assert x == CoproductTransform(y).swap()


def test_composite() -> None:
    x: Coproduct[int, Coproduct[str, bool]] = Coproduct.inl(23)
    y: Coproduct[Coproduct[int, bool], str] = Coproduct.inl(Coproduct.inl(23))
    f: CoproductFactory[int, Coproduct[bool, str]] = CoproductFactory()
    k: Coproduct[int, Coproduct[bool, str]] = x.map(
        lambda x: f.inl(x),
        lambda c: f.inr(CoproductTransform(c).swap()),
    )
    z = CoproductTransform.permutate(k)
    assert z == y



from ._core import (
    JsonObj,
    JsonValue,
)
from ._factory import (
    JsonValueFactory,
    UnfoldedFactory,
)
from ._transform import (
    Unfolder,
)

__all__ = [
    "JsonValue",
    "JsonObj",
    "JsonValueFactory",
    "UnfoldedFactory",
    "Unfolder",
]



from decimal import (
    Decimal,
)
from fa_purity.frozen import (
    freeze,
)
from fa_purity.json import (
    JsonValue as LegacyJsonValue,
)
from fa_purity.json.transform import (
    dumps as legacy_dumps,
)
from fa_purity.json_2 import (
    JsonPrimitiveFactory,
    JsonValue,
    Unfolder,
)


def test_legacy_json() -> None:
    obj = freeze({"foo": LegacyJsonValue(Decimal("123.44"))})
    assert legacy_dumps(obj)


def test_json_2() -> None:
    obj = freeze(
        {
            "foo": JsonValue.from_primitive(
                JsonPrimitiveFactory.from_raw(Decimal("123.44"))
            )
        }
    )
    assert Unfolder.dumps(obj)



from fa_purity.cmd import (
    Cmd,
)
from fa_purity.pure_iter.factory import (
    from_flist,
    from_range,
)
from fa_purity.pure_iter.transform import (
    chain,
    consume,
    filter_opt,
    until_none,
)
import pytest
from tests.pure_iter._utils import (
    assert_immutability,
)


def test_chain() -> None:
    base = (4, 78, 6)
    items = from_range(range(5)).map(lambda i: from_flist(base))
    chained = chain(items)
    assert_immutability(chained)
    assert tuple(chained) == base * 5


def test_consume() -> None:
    mutable_obj = [0]

    def _mutate(num: int) -> None:
        mutable_obj[0] = num

    items = from_flist(tuple(range(5))).map(
        lambda i: Cmd.from_cmd(lambda: _mutate(i))
    )
    assert_immutability(items, True)
    cmd = consume(items)
    assert mutable_obj[0] == 0

    def _verify(_: None) -> None:
        assert mutable_obj[0] == 4

    with pytest.raises(SystemExit):
        cmd.map(_verify).compute()


def test_filter_opt() -> None:
    items = tuple((0, 1, 2, 3, None, 4, 5, 6))
    result = filter_opt(from_flist(items))
    assert_immutability(result)
    assert tuple(result) == tuple(range(7))


def test_until_none() -> None:
    items = tuple((0, 1, 2, 3, None, 4, 5, 6))
    result = until_none(from_flist(items))
    assert_immutability(result)
    assert tuple(result) == tuple(range(4))



from fa_purity.maybe import (
    Maybe,
)
from fa_purity.pure_iter.core import (
    PureIter,
)
from fa_purity.pure_iter.factory import (
    from_flist,
    from_range,
)
from tests.pure_iter._utils import (
    assert_immutability,
)
from typing import (
    Callable,
)


def test_map() -> None:
    items = from_range(range(10)).map(lambda i: i + 1)
    assert_immutability(items)


def test_chunked() -> None:
    items = from_range(range(10)).chunked(2)
    assert_immutability(items)


def test_find_first() -> None:
    items = from_range(range(10))
    assert_immutability(items)
    assert items.find_first(lambda n: n > 6).unwrap() == 7
    assert items.find_first(lambda n: n > 99) == Maybe.empty()


def test_bind() -> None:
    items = from_range(range(5))
    items2: Callable[[int], PureIter[int]] = lambda n: from_flist((n, n))
    expected = (0, 0, 1, 1, 2, 2, 3, 3, 4, 4)
    result = items.bind(items2)
    assert_immutability(result)
    assert result.to_list() == expected



from __future__ import (
    annotations,
)

from deprecated import (
    deprecated,
)
from fa_purity.frozen import (
    FrozenDict,
    FrozenList,
)
from fa_purity.json.errors import (
    invalid_type,
)
from fa_purity.json.primitive.core import (
    Primitive,
)
from fa_purity.json.value import (
    factory as jval_factory,
)
from fa_purity.json.value.core import (
    JsonValue,
    UnfoldedJVal,
)
from fa_purity.json.value.transform import (
    Unfolder,
    UnfoldResult,
)
from fa_purity.result import (
    Result,
)
from fa_purity.utils import (
    raise_exception,
)
import simplejson
from typing import (
    Any,
    cast,
    Dict,
    IO as IO_FILE,
    TypeVar,
)

_T = TypeVar("_T")
JsonObj = FrozenDict[str, JsonValue]


class UnexpectedResult(Exception):
    pass


@deprecated("Replaced. Use `fa_purity.json_2.value.JsonPrimitiveFactory` instead")  # type: ignore[misc]
def from_any(raw: _T) -> UnfoldResult[JsonObj]:
    return jval_factory.from_any(raw).bind(lambda j: Unfolder(j).to_json())


@deprecated("NEW API: use `from_unfolded_dict` instead; function arg is mutable and not generic enough")  # type: ignore[misc]
def from_prim_dict(raw: Dict[str, Primitive]) -> JsonObj:
    return FrozenDict({key: JsonValue(val) for key, val in raw.items()})


@deprecated("Replaced. Use `fa_purity.json_2.UnfoldedFactory` instead")  # type: ignore[misc]
def from_unfolded_dict(raw: FrozenDict[str, UnfoldedJVal]) -> JsonObj:
    return FrozenDict({key: JsonValue(val) for key, val in raw.items()})


@deprecated("Replaced. Use `fa_purity.json_2.JsonValueFactory` instead")  # type: ignore[misc]
def loads(raw: str) -> UnfoldResult[JsonObj]:
    raw_json = cast(Dict[str, Any], simplejson.loads(raw))  # type: ignore[misc]
    return from_any(raw_json)  # type: ignore[misc]


@deprecated("Replaced. Use `fa_purity.json_2.JsonValueFactory` instead")  # type: ignore[misc]
def load(raw: IO_FILE[str]) -> UnfoldResult[JsonObj]:
    raw_json = cast(Dict[str, Any], simplejson.load(raw))  # type: ignore[misc]
    return from_any(raw_json)  # type: ignore[misc]


@deprecated("Replaced. Use `fa_purity.json_2.JsonValueFactory` instead")  # type: ignore[misc]
def json_list(raw: _T) -> UnfoldResult[FrozenList[JsonObj]]:
    try:
        if isinstance(raw, (list, tuple)):
            return Result.success(
                tuple(
                    from_any(item).alt(raise_exception).unwrap()
                    for item in raw
                )
            )
        return Result.failure(invalid_type.new("json_list", "List|Tuple", raw))
    except invalid_type.InvalidType as err:
        return Result.failure(err)



from __future__ import (
    annotations,
)

from dataclasses import (
    dataclass,
)
from deprecated import (
    deprecated,
)
from fa_purity._bug import (
    LibraryBug,
)
from typing import (
    Callable,
    Generic,
    Optional,
    Type,
    TypeVar,
    Union,
)

_A = TypeVar("_A")
_B = TypeVar("_B")
_C = TypeVar("_C")
_L = TypeVar("_L")
_R = TypeVar("_R")
_T = TypeVar("_T")


@deprecated("NEW API: use `UnionFactory` instead")  # type: ignore[misc]
def inr(val: _R, _left: Optional[Type[_L]] = None) -> Union[_L, _R]:
    return val


@deprecated("NEW API: use `UnionFactory` instead")  # type: ignore[misc]
def inl(val: _L, _right: Optional[Type[_R]] = None) -> Union[_L, _R]:
    return val


@dataclass(frozen=True)
class UnionFactory(Generic[_L, _R]):
    def inl(self, value: _L) -> Union[_L, _R]:
        return value

    def inr(self, value: _R) -> Union[_L, _R]:
        return value


@dataclass(frozen=True)
class _Empty:
    pass


@dataclass(frozen=True)
class _Coproduct(Generic[_L, _R]):
    left: Union[_L, _Empty]
    right: Union[_R, _Empty]
    left_val: bool

    @staticmethod
    def assert_non_empty(item: Union[_A, _Empty]) -> _A:
        if isinstance(item, _Empty):
            raise LibraryBug(
                Exception("assert_non_empty received empty input")
            )
        return item


@dataclass(frozen=True)
class Coproduct(Generic[_L, _R]):
    _inner: _Coproduct[_L, _R]

    @staticmethod
    def inl(value: _L) -> Coproduct[_L, _R]:
        return Coproduct(_Coproduct(value, _Empty(), True))

    @staticmethod
    def inr(value: _R) -> Coproduct[_L, _R]:
        return Coproduct(_Coproduct(_Empty(), value, False))

    def map(
        self, transform_1: Callable[[_L], _T], transform_2: Callable[[_R], _T]
    ) -> _T:
        if self._inner.left_val:
            return transform_1(_Coproduct.assert_non_empty(self._inner.left))
        return transform_2(_Coproduct.assert_non_empty(self._inner.right))


@dataclass(frozen=True)
class CoproductFactory(Generic[_L, _R]):
    def inl(self, value: _L) -> Coproduct[_L, _R]:
        return Coproduct.inl(value)

    def inr(self, value: _R) -> Coproduct[_L, _R]:
        return Coproduct.inr(value)


@dataclass(frozen=True)
class CoproductTransform(Generic[_L, _R]):
    _value: Coproduct[_L, _R]

    def swap(self) -> Coproduct[_R, _L]:
        def _right(item: _R) -> Coproduct[_R, _L]:
            return Coproduct.inl(item)

        def _left(item: _L) -> Coproduct[_R, _L]:
            return Coproduct.inr(item)

        return self._value.map(_left, _right)

    def to_union(self) -> Union[_L, _R]:
        factory: UnionFactory[_L, _R] = UnionFactory()
        return self._value.map(
            lambda l: factory.inl(l),
            lambda r: factory.inr(r),
        )

    @staticmethod
    def permutate(
        item: Coproduct[_A, Coproduct[_B, _C]]
    ) -> Coproduct[Coproduct[_A, _B], _C]:
        factory: CoproductFactory[Coproduct[_A, _B], _C] = CoproductFactory()
        return item.map(
            lambda a: factory.inl(Coproduct.inl(a)),
            lambda bc: bc.map(
                lambda b: factory.inl(Coproduct.inr(b)),
                lambda c: Coproduct.inr(c),
            ),
        )



from .core import (
    is_primitive,
    NotNonePrimTvar,
    Primitive,
    PrimitiveTVar,
    PrimitiveTypes,
    PrimitiveTypesList,
)

__all__ = [
    "Primitive",
    "PrimitiveTypes",
    "PrimitiveTypesList",
    "PrimitiveTVar",
    "NotNonePrimTvar",
    "is_primitive",
]



from fa_purity import (
    _iter_factory,
)
from fa_purity.cmd import (
    Cmd,
)
from fa_purity.frozen import (
    FrozenList,
)
from fa_purity.pure_iter._inner import (
    InnerPureIter,
)
from fa_purity.pure_iter.core import (
    PureIter,
)
from itertools import (
    count,
)
from typing import (
    Callable,
    Iterable,
    List,
    TypeVar,
    Union,
)

_T = TypeVar("_T")
_I = TypeVar("_I")
_R = TypeVar("_R")


def unsafe_from_cmd(cmd: Cmd[Iterable[_T]]) -> PureIter[_T]:
    # [WARNING] unsafe constructor
    # - Type-check cannot ensure its proper use
    # - Do not use until is strictly necessary
    # - Do unit test over the function defined by this, i.e. assert reproducibility
    #
    # Cmd argument MUST produce semanticly equivalent iterables. This is:
    # possibly different objects that means the same thing
    #
    # - if Iterable is IMMUTABLE (e.g. tuple) then requirement is fulfilled
    # - if Iterable is MUTABLE then the Cmd must call the obj constructor (that is not pure)
    # with the same arguments for ensuring equivalence.
    #
    # Non compliant code:
    #   y = map(lambda i: i + 1, range(0, 10))
    #   x = unsafe_from_cmd(
    #       Cmd.from_cmd(lambda: y)
    #   )
    #   # y is a map obj instance; cmd lambda is pinned with a single ref
    #   # since map is MUTABLE the ref should change at every call
    #
    # Compliant code:
    #   x = unsafe_from_cmd(
    #       Cmd.from_cmd(
    #           lambda: map(lambda i: i + 1, range(0, 10))
    #       )
    #   )
    #   # cmd lambda produces a new ref in each call
    #   # but all of them are equivalent (created with the same args)
    return PureIter(InnerPureIter(cmd))


def from_flist(items: FrozenList[_T]) -> PureIter[_T]:
    return unsafe_from_cmd(Cmd.from_cmd(lambda: items))


def from_list(items: Union[List[_T], FrozenList[_T]]) -> PureIter[_T]:
    _items = tuple(items) if isinstance(items, list) else items
    return from_flist(_items)


def from_range(range_obj: range) -> PureIter[int]:
    return unsafe_from_cmd(Cmd.from_cmd(lambda: range_obj))


def infinite_gen(function: Callable[[_T], _T], init: _T) -> PureIter[_T]:
    return unsafe_from_cmd(
        Cmd.from_cmd(lambda: _iter_factory.infinite_gen(function, init))
    )


def infinite_range(start: int, step: int) -> PureIter[int]:
    return unsafe_from_cmd(Cmd.from_cmd(lambda: count(start, step)))


def pure_map(
    function: Callable[[_I], _R], items: Union[List[_I], FrozenList[_I]]
) -> PureIter[_R]:
    return from_list(items).map(function)



from fa_purity.cmd import (
    unsafe_unwrap,
)
from fa_purity.cmd.core import (
    Cmd,
)
from fa_purity.maybe import (
    Maybe,
)
from fa_purity.pure_iter.factory import (
    from_flist,
    from_range,
)
from fa_purity.stream.core import (
    Stream,
)
from fa_purity.stream.factory import (
    from_piter,
)
from tests.stream._utils import (
    assert_different_iter,
    rand_int,
)
from typing import (
    Callable,
)


def test_map() -> None:
    items = from_range(range(10)).map(lambda _: rand_int())
    stm = from_piter(items).map(lambda i: i + 1)
    assert_different_iter(stm)


def test_chunked() -> None:
    items = from_range(range(10)).map(lambda _: rand_int())
    stm = from_piter(items).chunked(2)
    assert_different_iter(stm)


def test_find_first() -> None:
    items = from_piter(from_range(range(10)).map(lambda _: rand_int()))
    assert_different_iter(items)
    assert unsafe_unwrap(items.find_first(lambda n: n >= 0)).unwrap() >= 0
    assert unsafe_unwrap(items.find_first(lambda n: n > 9999)) == Maybe.empty()


def test_bind() -> None:
    items = from_piter(
        from_range(range(5)).map(lambda i: Cmd.from_cmd(lambda: i))
    )
    items2: Callable[[int], Stream[int]] = lambda n: from_piter(
        from_flist((n, n)).map(lambda i: Cmd.from_cmd(lambda: i))
    )
    expected = (0, 0, 1, 1, 2, 2, 3, 3, 4, 4)
    result = items.bind(items2)
    assert_different_iter(result)
    assert unsafe_unwrap(result.to_list()) == expected



from __future__ import (
    annotations,
)

from dataclasses import (
    dataclass,
    field,
)
from decimal import (
    Decimal,
)
from fa_purity.union import (
    Coproduct,
)
from typing import (
    Callable,
    TypeVar,
)

_T = TypeVar("_T")

_bool_or_none = Coproduct[bool, None]
_decimal_or = Coproduct[Decimal, _bool_or_none]
_float_or = Coproduct[float, _decimal_or]
_int_or = Coproduct[int, _float_or]
_JsonPrimitive = Coproduct[str, _int_or]


@dataclass(frozen=True)
class _Private:
    pass


@dataclass(frozen=True)
class JsonPrimitive:
    _private: _Private = field(repr=False, hash=False, compare=False)
    _value: _JsonPrimitive

    def map(
        self,
        str_case: Callable[[str], _T],
        int_case: Callable[[int], _T],
        float_case: Callable[[float], _T],
        decimal_case: Callable[[Decimal], _T],
        bool_case: Callable[[bool], _T],
        none_case: Callable[[], _T],
    ) -> _T:
        return self._value.map(
            str_case,
            lambda a: a.map(
                int_case,
                lambda b: b.map(
                    float_case,
                    lambda c: c.map(
                        decimal_case,
                        lambda d: d.map(bool_case, lambda _: none_case()),
                    ),
                ),
            ),
        )

    @staticmethod
    def from_str(item: str) -> JsonPrimitive:
        return JsonPrimitive(_Private(), Coproduct.inl(item))

    @staticmethod
    def from_int(item: int) -> JsonPrimitive:
        return JsonPrimitive(_Private(), Coproduct.inr(Coproduct.inl(item)))

    @staticmethod
    def from_float(item: float) -> JsonPrimitive:
        return JsonPrimitive(
            _Private(), Coproduct.inr(Coproduct.inr(Coproduct.inl(item)))
        )

    @staticmethod
    def from_decimal(item: Decimal) -> JsonPrimitive:
        return JsonPrimitive(
            _Private(),
            Coproduct.inr(Coproduct.inr(Coproduct.inr(Coproduct.inl(item)))),
        )

    @staticmethod
    def from_bool(item: bool) -> JsonPrimitive:
        return JsonPrimitive(
            _Private(),
            Coproduct.inr(
                Coproduct.inr(
                    Coproduct.inr(Coproduct.inr(Coproduct.inl(item)))
                )
            ),
        )

    @staticmethod
    def empty() -> JsonPrimitive:
        return JsonPrimitive(
            _Private(),
            Coproduct.inr(
                Coproduct.inr(
                    Coproduct.inr(Coproduct.inr(Coproduct.inr(None)))
                )
            ),
        )



from ._core import (
    JsonPrimitive,
)
from dataclasses import (
    dataclass,
)
from decimal import (
    Decimal,
)
from fa_purity.result import (
    ResultE,
)
from fa_purity.result.core import (
    Result,
)
from fa_purity.union import (
    UnionFactory,
)
from typing import (
    Optional,
)


@dataclass(frozen=True)
class JsonPrimitiveUnfolder:
    @staticmethod
    def to_str(item: JsonPrimitive) -> ResultE[str]:
        fail: ResultE[str] = Result.failure(
            TypeError("Unfolded JsonPrimitive is not `str`"), str
        ).alt(Exception)
        return item.map(
            lambda x: Result.success(x),
            lambda _: fail,
            lambda _: fail,
            lambda _: fail,
            lambda _: fail,
            lambda: fail,
        )

    @staticmethod
    def to_int(item: JsonPrimitive) -> ResultE[int]:
        fail: ResultE[int] = Result.failure(
            TypeError("Unfolded JsonPrimitive is not `int`"), int
        ).alt(Exception)
        return item.map(
            lambda _: fail,
            lambda x: Result.success(x),
            lambda _: fail,
            lambda _: fail,
            lambda _: fail,
            lambda: fail,
        )

    @staticmethod
    def to_float(item: JsonPrimitive) -> ResultE[float]:
        fail: ResultE[float] = Result.failure(
            TypeError("Unfolded JsonPrimitive is not `float`"), float
        ).alt(Exception)
        return item.map(
            lambda _: fail,
            lambda _: fail,
            lambda x: Result.success(x),
            lambda _: fail,
            lambda _: fail,
            lambda: fail,
        )

    @staticmethod
    def to_decimal(item: JsonPrimitive) -> ResultE[Decimal]:
        fail: ResultE[Decimal] = Result.failure(
            TypeError("Unfolded JsonPrimitive is not `Decimal`"), Decimal
        ).alt(Exception)
        return item.map(
            lambda _: fail,
            lambda _: fail,
            lambda _: fail,
            lambda x: Result.success(x),
            lambda _: fail,
            lambda: fail,
        )

    @staticmethod
    def to_bool(item: JsonPrimitive) -> ResultE[bool]:
        fail: ResultE[bool] = Result.failure(
            TypeError("Unfolded JsonPrimitive is not `bool`"), bool
        ).alt(Exception)
        return item.map(
            lambda _: fail,
            lambda _: fail,
            lambda _: fail,
            lambda _: fail,
            lambda x: Result.success(x),
            lambda: fail,
        )

    @staticmethod
    def to_none(item: JsonPrimitive) -> ResultE[None]:
        fail: ResultE[None] = Result.failure(
            TypeError("Unfolded JsonPrimitive is not `None`"), type(None)
        ).alt(Exception)
        return item.map(
            lambda _: fail,
            lambda _: fail,
            lambda _: fail,
            lambda _: fail,
            lambda _: fail,
            lambda: Result.success(None),
        )

    @classmethod
    def to_opt_str(cls, item: JsonPrimitive) -> ResultE[Optional[str]]:
        factory: UnionFactory[str, None] = UnionFactory()
        return (
            cls.to_str(item)
            .map(factory.inl)
            .lash(lambda _: cls.to_none(item).map(factory.inr))
        )

    @classmethod
    def to_opt_int(cls, item: JsonPrimitive) -> ResultE[Optional[int]]:
        factory: UnionFactory[int, None] = UnionFactory()
        return (
            cls.to_int(item)
            .map(factory.inl)
            .lash(lambda _: cls.to_none(item).map(factory.inr))
        )

    @classmethod
    def to_opt_float(cls, item: JsonPrimitive) -> ResultE[Optional[float]]:
        factory: UnionFactory[float, None] = UnionFactory()
        return (
            cls.to_float(item)
            .map(factory.inl)
            .lash(lambda _: cls.to_none(item).map(factory.inr))
        )

    @classmethod
    def to_opt_decimal(cls, item: JsonPrimitive) -> ResultE[Optional[Decimal]]:
        factory: UnionFactory[Decimal, None] = UnionFactory()
        return (
            cls.to_decimal(item)
            .map(factory.inl)
            .lash(lambda _: cls.to_none(item).map(factory.inr))
        )

    @classmethod
    def to_opt_bool(cls, item: JsonPrimitive) -> ResultE[Optional[bool]]:
        factory: UnionFactory[bool, None] = UnionFactory()
        return (
            cls.to_bool(item)
            .map(factory.inl)
            .lash(lambda _: cls.to_none(item).map(factory.inr))
        )



from fa_purity.cmd import (
    Cmd,
)
from fa_purity.cmd.transform import (
    serial_merge,
)
from fa_purity.frozen import (
    FrozenList,
)
from fa_purity.result import (
    Result,
)
from fa_purity.result.transform import (
    all_ok,
)
from typing import (
    TypeVar,
)

_T = TypeVar("_T")
_S = TypeVar("_S")
_F = TypeVar("_F")


def flatten_cmds(items: FrozenList[Cmd[_T]]) -> Cmd[FrozenList[_T]]:
    """serial_merge function alias"""
    return serial_merge(items)


def flatten_results(
    items: FrozenList[Result[_S, _F]]
) -> Result[FrozenList[_S], _F]:
    """all_ok function alias"""
    return all_ok(items)



from ._inner import (
    InnerStream,
)
from fa_purity import (
    _iter_factory,
)
from fa_purity.cmd import (
    Cmd,
    unsafe_unwrap,
)
from fa_purity.maybe import (
    Maybe,
)
from fa_purity.pure_iter.core import (
    PureIter,
)
from fa_purity.stream.core import (
    Stream,
)
from fa_purity.stream.factory import (
    from_piter,
)
from typing import (
    Optional,
    TypeVar,
    Union,
)

_T = TypeVar("_T")


def _chain_1(
    unchained: Stream[PureIter[_T]],
) -> Stream[_T]:
    return unchained.map(
        lambda p: p.map(lambda x: Cmd.from_cmd(lambda: x)).transform(
            lambda x: from_piter(x)
        )
    ).bind(lambda x: x)


def _chain_2(
    unchained: PureIter[Stream[_T]],
) -> Stream[_T]:
    return (
        unchained.map(lambda s: Cmd.from_cmd(lambda: s))
        .transform(lambda x: from_piter(x))
        .bind(lambda x: x)
    )


def chain(
    unchained: Union[Stream[PureIter[_T]], PureIter[Stream[_T]]],
) -> Stream[_T]:
    if isinstance(unchained, Stream):
        return _chain_1(unchained)
    return _chain_2(unchained)


def squash(stm: Stream[Cmd[_T]]) -> Stream[_T]:
    items = InnerStream(stm.unsafe_to_iter().bind(_iter_factory.squash))
    return Stream(items)


def consume(stm: Stream[Cmd[None]]) -> Cmd[None]:
    return Cmd.from_cmd(
        lambda: _iter_factory.deque(
            iter(unsafe_unwrap(a) for a in unsafe_unwrap(stm.unsafe_to_iter()))
        )
    )


def filter_opt(stm: Stream[Optional[_T]]) -> Stream[_T]:
    items = InnerStream(stm.unsafe_to_iter().map(_iter_factory.filter_none))
    return Stream(items)


def filter_maybe(stm: Stream[Maybe[_T]]) -> Stream[_T]:
    return filter_opt(stm.map(lambda x: x.value_or(None)))


def until_none(stm: Stream[Optional[_T]]) -> Stream[_T]:
    items = InnerStream(stm.unsafe_to_iter().map(_iter_factory.until_none))
    return Stream(items)


def until_empty(stm: Stream[Maybe[_T]]) -> Stream[_T]:
    return until_none(stm.map(lambda m: m.value_or(None)))



from dataclasses import (
    dataclass,
)
from deprecated import (
    deprecated,
)
from fa_purity.frozen import (
    freeze,
)
from fa_purity.json import (
    JsonObj as LegacyJsonObj,
    JsonValue as LegacyJsonValue,
)
from fa_purity.json_2.primitive import (
    JsonPrimitive,
    JsonPrimitiveFactory,
    JsonPrimitiveUnfolder,
    Primitive,
)
from fa_purity.json_2.value import (
    JsonObj,
    JsonValue,
    JsonValueFactory,
    UnfoldedFactory,
    Unfolder,
)


@deprecated("Temporal adapter for `json` -> `json_2` migration")  # type: ignore[misc]
@dataclass(frozen=True)
class LegacyAdapter:
    @classmethod
    def json_value(cls, legacy: LegacyJsonValue) -> JsonValue:
        return legacy.map(
            lambda p: JsonValue.from_primitive(
                JsonPrimitiveFactory.from_raw(p)
            ),
            lambda items: JsonValue.from_list(
                tuple(cls.json_value(i) for i in items)
            ),
            lambda d: JsonValue.from_json(
                freeze({k: cls.json_value(v) for k, v in d.items()})
            ),
        )

    @classmethod
    def json(cls, legacy: LegacyJsonObj) -> JsonObj:
        _new = {k: cls.json_value(v) for k, v in legacy.items()}
        return freeze(_new)

    @classmethod
    def to_legacy_value(cls, value: JsonValue) -> LegacyJsonValue:
        return value.map(
            lambda p: p.map(
                lambda x: LegacyJsonValue(x),
                lambda x: LegacyJsonValue(x),
                lambda x: LegacyJsonValue(x),
                lambda x: LegacyJsonValue(x),
                lambda x: LegacyJsonValue(x),
                lambda: LegacyJsonValue(None),
            ),
            lambda items: LegacyJsonValue(
                tuple(cls.to_legacy_value(i) for i in items)
            ),
            lambda d: LegacyJsonValue(
                freeze({k: cls.to_legacy_value(v) for k, v in d.items()})
            ),
        )

    @classmethod
    def to_legacy_json(cls, value: JsonObj) -> LegacyJsonObj:
        _new = {k: cls.to_legacy_value(v) for k, v in value.items()}
        return freeze(_new)


__all__ = [
    "Primitive",
    "JsonPrimitive",
    "JsonPrimitiveFactory",
    "JsonPrimitiveUnfolder",
    "JsonValue",
    "JsonObj",
    "JsonValueFactory",
    "UnfoldedFactory",
    "Unfolder",
]



from .core import (
    Result,
    ResultE,
)
from dataclasses import (
    dataclass,
)
from fa_purity.frozen import (
    FrozenDict,
)
from typing import (
    Generic,
    TypeVar,
)

_K = TypeVar("_K")
_V = TypeVar("_V")
_S = TypeVar("_S")
_F = TypeVar("_F")


@dataclass(frozen=True)
class ResultFactory(Generic[_S, _F]):
    def success(self, value: _S) -> Result[_S, _F]:
        return Result.success(value)

    def failure(self, value: _F) -> Result[_S, _F]:
        return Result.failure(value)


def try_get(data: FrozenDict[_K, _V], key: _K) -> ResultE[_V]:
    factory: ResultFactory[_V, Exception] = ResultFactory()
    if key in data:
        return factory.success(data[key])
    return factory.failure(KeyError(key)).alt(Exception)



from __future__ import (
    annotations,
)

from dataclasses import (
    dataclass,
)
from fa_purity.union import (
    Coproduct,
    CoproductFactory,
    CoproductTransform,
)
from fa_purity.utils import (
    raise_exception,
)
from typing import (
    Callable,
    Generic,
    NoReturn,
    Optional,
    Type,
    TypeVar,
    Union,
)

_S = TypeVar("_S")
_F = TypeVar("_F")
_T = TypeVar("_T")


@dataclass(frozen=True)
class UnwrapError(Exception, Generic[_S, _F]):
    # DO NOT catch specific `UnwrapError[_S, _F]` error
    # catch always `UnwrapError`
    container: "Result[_S, _F]"


@dataclass(frozen=True)
class _Result(Generic[_S, _F]):
    value: Coproduct[_S, _F]


@dataclass(frozen=True)
class Result(Generic[_S, _F]):
    _inner: _Result[_S, _F]

    @staticmethod
    def success(val: _S, _type: Optional[Type[_F]] = None) -> Result[_S, _F]:
        item: Coproduct[_S, _F] = Coproduct.inl(val)
        return Result(_Result(item))

    @staticmethod
    def failure(val: _F, _type: Optional[Type[_S]] = None) -> Result[_S, _F]:
        item: Coproduct[_S, _F] = Coproduct.inr(val)
        return Result(_Result(item))

    def map(self, function: Callable[[_S], _T]) -> Result[_T, _F]:
        factory: CoproductFactory[_T, _F] = CoproductFactory()
        val = self._inner.value.map(
            lambda s: factory.inl(function(s)),
            lambda f: factory.inr(f),
        )
        return Result(_Result(val))

    def alt(self, function: Callable[[_F], _T]) -> Result[_S, _T]:
        factory: CoproductFactory[_S, _T] = CoproductFactory()
        val = self._inner.value.map(
            lambda s: factory.inl(s),
            lambda f: factory.inr(function(f)),
        )
        return Result(_Result(val))

    def to_coproduct(self) -> Coproduct[_S, _F]:
        return self._inner.value

    def bind(self, function: Callable[[_S], Result[_T, _F]]) -> Result[_T, _F]:
        factory: CoproductFactory[_T, _F] = CoproductFactory()
        val = self._inner.value.map(
            lambda s: function(s).to_coproduct(),
            lambda f: factory.inr(f),
        )
        return Result(_Result(val))

    def lash(self, function: Callable[[_F], Result[_S, _T]]) -> Result[_S, _T]:
        factory: CoproductFactory[_S, _T] = CoproductFactory()
        val = self._inner.value.map(
            lambda s: factory.inl(s),
            lambda f: function(f).to_coproduct(),
        )
        return Result(_Result(val))

    def swap(self) -> Result[_F, _S]:
        val = CoproductTransform(self._inner.value).swap()
        return Result(_Result(val))

    def apply(self, wrapped: Result[Callable[[_S], _T], _F]) -> Result[_T, _F]:
        return wrapped.bind(lambda f: self.map(f))

    def cop_value_or(self, default: _T) -> Coproduct[_S, _T]:
        factory: CoproductFactory[_S, _T] = CoproductFactory()
        val = self._inner.value.map(
            lambda s: factory.inl(s),
            lambda _: factory.inr(default),
        )
        return val

    def cop_or_else_call(
        self, function: Callable[[], _T]
    ) -> Coproduct[_S, _T]:
        factory: CoproductFactory[_S, _T] = CoproductFactory()
        val = self._inner.value.map(
            lambda s: factory.inl(s),
            lambda _: factory.inr(function()),
        )
        return val

    def value_or(self, default: _T) -> Union[_S, _T]:
        return CoproductTransform(self.cop_value_or(default)).to_union()

    def or_else_call(self, function: Callable[[], _T]) -> Union[_S, _T]:
        return CoproductTransform(self.cop_or_else_call(function)).to_union()

    def to_union(self) -> Union[_S, _F]:
        return CoproductTransform(self._inner.value).to_union()

    def unwrap(self) -> Union[_S, NoReturn]:
        val = self._inner.value.map(
            lambda s: s,
            lambda _: raise_exception(UnwrapError(self)),
        )
        return val

    def unwrap_fail(self) -> Union[_F, NoReturn]:
        val = self._inner.value.map(
            lambda _: raise_exception(UnwrapError(self)),
            lambda f: f,
        )
        return val


ResultE = Result[_T, Exception]  # type: ignore[misc]



from fa_purity.pure_iter.factory import (
    from_range,
    infinite_range,
)
from tests.pure_iter._utils import (
    assert_immutability,
    assert_immutability_inf,
)


def test_use_case_1() -> None:
    items = from_range(range(0, 10))
    mapped = items.map(lambda i: i + 2)
    assert mapped.to_list() == tuple(range(2, 12))
    r = items.map(lambda x: x * 2).chunked(3)
    expected = ((0, 2, 4), (6, 8, 10), (12, 14, 16), (18,))
    assert r.to_list() == expected
    assert_immutability(items)
    assert_immutability(mapped)


def test_inf() -> None:
    items = infinite_range(4, 10)
    for n, v in enumerate(items):
        assert v == 4 + (n * 10)
        if n > 15:
            break
    assert_immutability_inf(items)



from __future__ import (
    annotations,
)

from decimal import (
    Decimal,
)
from deprecated import (
    deprecated,
)
from typing import (
    Type,
    TypeVar,
    Union,
)
from typing_extensions import (
    TypeGuard,
)

_T = TypeVar("_T")
Primitive = Union[str, int, float, Decimal, bool, None]
PrimitiveTypes = Union[
    Type[str],
    Type[int],
    Type[float],
    Type[Decimal],
    Type[bool],
    Type[None],
]
PrimitiveTypesList = (
    str,
    int,
    float,
    Decimal,
    bool,
    type(None),
)
PrimitiveTVar = TypeVar(  # Deprecated use NotNonePrimTvar instead
    "PrimitiveTVar", str, int, float, Decimal, bool, Type[None]
)
NotNonePrimTvar = TypeVar("NotNonePrimTvar", str, int, float, Decimal, bool)


@deprecated("Replaced. Use `fa_purity.json_2.primitive.JsonPrimitive` instead")  # type: ignore[misc]
def is_primitive(raw: _T) -> TypeGuard[Primitive]:
    if raw is None or isinstance(raw, PrimitiveTypesList):
        return True
    return False



from deprecated import (
    deprecated,
)
from typing import (
    TypeVar,
)

_T = TypeVar("_T")


class _InvalidType(Exception):
    pass


@deprecated("Replaced with generic Exception")  # type: ignore[misc]
class InvalidType(_InvalidType):
    def __init__(self, obj: _InvalidType) -> None:
        super().__init__(obj)


@deprecated("Replaced with generic Exception")  # type: ignore[misc]
def new(caller: str, expected: str, item: _T) -> InvalidType:
    draft = _InvalidType(
        f"{caller} expected `{expected}` not `{str(type(item))}`"
    )
    return InvalidType(draft)



from deprecated import (
    deprecated,
)
from fa_purity.frozen import (
    FrozenDict,
    FrozenList,
)
from fa_purity.json.errors import (
    invalid_type,
)
from fa_purity.json.errors.invalid_type import (
    InvalidType,
)
from fa_purity.json.primitive import (
    factory as prim_factory,
)
from fa_purity.json.primitive.core import (
    is_primitive,
    Primitive,
)
from fa_purity.json.value.core import (
    JsonValue,
)
from fa_purity.result import (
    Result,
)
from fa_purity.utils import (
    raise_exception,
)
from typing import (
    Dict,
    List,
    TypeVar,
    Union,
)

_T = TypeVar("_T")


@deprecated("Replaced. Use `fa_purity.json_2.value.JsonValueFactory` methods instead")  # type: ignore[misc]
def from_list(raw: Union[List[Primitive], FrozenList[Primitive]]) -> JsonValue:
    return JsonValue(tuple(JsonValue(item) for item in raw))


@deprecated("Replaced. Use `fa_purity.json_2.value.JsonValueFactory` methods instead")  # type: ignore[misc]
def from_dict(
    raw: Union[Dict[str, Primitive], FrozenDict[str, Primitive]]
) -> JsonValue:
    return JsonValue(
        FrozenDict({key: JsonValue(val) for key, val in raw.items()})
    )


@deprecated("Replaced. Use `fa_purity.json_2.value.JsonValueFactory` methods instead")  # type: ignore[misc]
def from_any(raw: _T) -> Result[JsonValue, InvalidType]:
    if is_primitive(raw):
        return Result.success(JsonValue(raw))
    if isinstance(raw, (FrozenDict, dict)):
        try:
            json_dict = FrozenDict(
                {
                    prim_factory.to_primitive(key, str)
                    .alt(raise_exception)
                    .unwrap(): from_any(val)
                    .alt(raise_exception)
                    .unwrap()
                    for key, val in raw.items()
                }
            )
            return Result.success(JsonValue(json_dict))
        except InvalidType as err:
            return Result.failure(err)
    if isinstance(raw, (list, tuple)):
        try:
            json_list = tuple(
                from_any(item).alt(raise_exception).unwrap() for item in raw
            )
            return Result.success(JsonValue(json_list))
        except InvalidType as err:
            return Result.failure(err)
    return Result.failure(invalid_type.new("from_any", "UnfoldedJVal", raw))



from fa_purity import (
    _iter_factory,
)
from fa_purity.cmd import (
    Cmd,
)
from fa_purity.cmd.core import (
    CmdUnwrapper,
    new_cmd,
)
from fa_purity.maybe import (
    Maybe,
)
from fa_purity.pure_iter.core import (
    PureIter,
)
from fa_purity.pure_iter.factory import (
    unsafe_from_cmd,
)
from typing import (
    Optional,
    TypeVar,
)

_T = TypeVar("_T")


def chain(
    unchained: PureIter[PureIter[_T]],
) -> PureIter[_T]:
    return unchained.bind(lambda x: x)


def consume(p_iter: PureIter[Cmd[None]]) -> Cmd[None]:
    def _action(unwrapper: CmdUnwrapper) -> None:
        for c in p_iter:
            unwrapper.act(c)

    return Cmd.new_cmd(_action)


def filter_opt(items: PureIter[Optional[_T]]) -> PureIter[_T]:
    return unsafe_from_cmd(
        Cmd.from_cmd(lambda: _iter_factory.filter_none(items))
    )


def filter_maybe(items: PureIter[Maybe[_T]]) -> PureIter[_T]:
    return filter_opt(items.map(lambda x: x.value_or(None)))


def until_none(items: PureIter[Optional[_T]]) -> PureIter[_T]:
    return unsafe_from_cmd(
        Cmd.from_cmd(lambda: _iter_factory.until_none(items))
    )


def until_empty(items: PureIter[Maybe[_T]]) -> PureIter[_T]:
    return until_none(items.map(lambda m: m.value_or(None)))



from fa_purity.cmd import (
    Cmd,
)
from fa_purity.frozen import (
    FrozenList,
)
from fa_purity.pure_iter.factory import (
    infinite_range,
)
from fa_purity.stream.factory import (
    from_piter,
)
from fa_purity.stream.transform import (
    until_none,
)
import pytest
from secrets import (
    randbelow,
)
from typing import (
    Optional,
)


def _rand_val(count: int, none_index: int) -> Cmd[Optional[int]]:
    return Cmd.from_cmd(lambda: randbelow(11) if count != none_index else None)


def test_use_case_1() -> None:
    with pytest.raises(SystemExit):
        none_index = 7
        data = (
            infinite_range(0, 1)
            .map(lambda x: _rand_val(x, none_index))
            .transform(lambda x: from_piter(x))
        )
        result = data.transform(lambda x: until_none(x)).map(lambda n: n * -1)

        def _verify(items: FrozenList[int]) -> None:
            assert len(items) == none_index
            for n in items:
                assert n <= 0

        verification = result.to_list().map(_verify)
        verification.compute()



from dataclasses import (
    dataclass,
)
from fa_purity.cmd import (
    Cmd,
)
from typing import (
    Generic,
    Iterable,
    TypeVar,
)

_T = TypeVar("_T")


@dataclass(frozen=True)
class InnerStream(
    Generic[_T],
):
    new_iter: Cmd[Iterable[_T]]



from datetime import (
    timedelta,
    timezone,
)
from fa_purity.date_time import (
    DatetimeFactory,
)


def test_epoch() -> None:
    assert DatetimeFactory.EPOCH_START


def test_utc_addition() -> None:
    delta = timedelta(1, 2, 3, 5, 6, 7)
    date_time = DatetimeFactory.new_utc(2000, 1, 1, 0, 0, 0, 0)
    assert (date_time + delta).date_time == date_time.date_time + delta


def test_tz_addition() -> None:
    delta = timedelta(1, 2, 3, 5, 6, 7)
    date_time = DatetimeFactory.new_tz(2000, 1, 1, 0, 0, 0, 0, timezone.utc)
    assert (date_time + delta).date_time == date_time.date_time + delta



from fa_purity.json_2.value import (
    JsonValueFactory,
    Unfolder,
)

test_data = (
    JsonValueFactory.from_any({"foo": {"nested": ["hi", 99]}})
    .bind(Unfolder.to_json)
    .unwrap()
)


def test_dumps() -> None:
    assert Unfolder.dumps(test_data).replace(
        " ", ""
    ) == '{"foo": {"nested": ["hi", 99]} }'.replace(" ", "")



from __future__ import (
    annotations,
)

from ._inner import (
    InnerStream,
)
from dataclasses import (
    dataclass,
)
from fa_purity import (
    _iter_factory,
)
from fa_purity.cmd import (
    Cmd,
)
from fa_purity.frozen import (
    FrozenList,
)
from fa_purity.maybe import (
    Maybe,
)
import functools
from typing import (
    Callable,
    Generic,
    Iterable,
    TypeVar,
)

_T = TypeVar("_T")
_R = TypeVar("_R")


@dataclass(frozen=True)
class Stream(Generic[_T]):
    _inner: InnerStream[_T]

    def map(self, function: Callable[[_T], _R]) -> Stream[_R]:
        items: Cmd[Iterable[_R]] = self._inner.new_iter.map(
            lambda i: iter(map(function, i))
        )
        return Stream(InnerStream(items))

    def reduce(self, function: Callable[[_R, _T], _R], init: _R) -> Cmd[_R]:
        return self._inner.new_iter.map(
            lambda i: functools.reduce(function, i, init)
        )

    def bind(self, function: Callable[[_T], Stream[_R]]) -> Stream[_R]:
        items = (
            self.map(function)
            .unsafe_to_iter()
            .map(lambda items: iter(i.unsafe_to_iter() for i in items))
            .bind(lambda x: _iter_factory.squash(x))
            .map(lambda x: _iter_factory.chain(x))
        )
        return Stream(InnerStream(items))

    def filter(self, function: Callable[[_T], bool]) -> Stream[_T]:
        items: Cmd[Iterable[_T]] = self._inner.new_iter.map(
            lambda i: iter(filter(function, i))
        )
        return Stream(InnerStream(items))

    def find_first(self, criteria: Callable[[_T], bool]) -> Cmd[Maybe[_T]]:
        return self._inner.new_iter.map(
            lambda i: _iter_factory.find_first(criteria, i)
        ).map(lambda x: Maybe.from_optional(x))

    def chunked(self, size: int) -> Stream[FrozenList[_T]]:
        items: Cmd[Iterable[FrozenList[_T]]] = self._inner.new_iter.map(
            lambda i: _iter_factory.chunked(i, size)
        )
        return Stream(InnerStream(items))

    def transform(self, function: Callable[[Stream[_T]], _R]) -> _R:
        return function(self)

    def to_list(self) -> Cmd[FrozenList[_T]]:
        return self._inner.new_iter.map(tuple)

    def unsafe_to_iter(self) -> Cmd[Iterable[_T]]:
        # if possible iterables should not be used directly
        return self._inner.new_iter



from fa_purity.cmd import (
    Cmd,
    unsafe_unwrap,
)
from fa_purity.stream.core import (
    Stream,
)
from secrets import (
    randbelow,
)
from typing import (
    TypeVar,
)

_T = TypeVar("_T")


def rand_int() -> Cmd[int]:
    return Cmd.from_cmd(lambda: randbelow(11))


def assert_different_iter(stm: Stream[_T]) -> None:
    iter1 = unsafe_unwrap(stm.unsafe_to_iter())
    iter2 = unsafe_unwrap(stm.unsafe_to_iter())
    assert id(iter1) != id(iter2)



from fa_purity.pure_iter.factory import (
    from_range,
)
from fa_purity.stream.factory import (
    from_piter,
)
from tests.stream._utils import (
    assert_different_iter,
    rand_int,
)


def test_from_piter() -> None:
    items = from_range(range(10)).map(lambda _: rand_int())
    stm = from_piter(items)
    assert_different_iter(stm)



from dataclasses import (
    dataclass,
)
from typing import (
    Dict,
    Generic,
    Iterator,
    List,
    Mapping,
    overload,
    Tuple,
    TypeVar,
    Union,
)

_K = TypeVar("_K")
_V = TypeVar("_V")
_T = TypeVar("_T")

FrozenList = Tuple[_T, ...]  # type: ignore[misc]


@dataclass(frozen=True)
class _FrozenDict(Generic[_K, _V]):
    _dict: Dict[_K, _V]


class FrozenDict(Mapping[_K, _V], _FrozenDict[_K, _V]):
    def __init__(self, dictionary: Dict[_K, _V]):
        super().__init__(dictionary.copy())

    def __getitem__(self, key: _K) -> _V:
        return self._dict[key]

    def __iter__(self) -> Iterator[_K]:
        return iter(self._dict)

    def __len__(self) -> int:
        return len(self._dict)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}{self._dict}"

    def __hash__(self) -> int:
        return hash(frozenset(self._dict.items()))


def chain(unchained: FrozenList[FrozenList[_T]]) -> FrozenList[_T]:
    return tuple(item for items in unchained for item in items)


@overload
def freeze(target: List[_T]) -> FrozenList[_T]:
    pass


@overload
def freeze(target: Dict[_K, _V]) -> FrozenDict[_K, _V]:
    pass


def freeze(
    target: Union[List[_T], Dict[_K, _V]]
) -> Union[FrozenList[_T], FrozenDict[_K, _V]]:
    if isinstance(target, list):
        return tuple(target)
    return FrozenDict(target)


@overload
def unfreeze(target: FrozenList[_T]) -> List[_T]:
    pass


@overload
def unfreeze(target: FrozenDict[_K, _V]) -> Dict[_K, _V]:
    pass


def unfreeze(
    target: Union[FrozenList[_T], FrozenDict[_K, _V]]
) -> Union[List[_T], Dict[_K, _V]]:
    if isinstance(target, FrozenDict):
        return {k: v for k, v in target.items()}
    return [i for i in target]



from fa_purity.frozen import (
    FrozenList,
)
from fa_purity.pure_iter.core import (
    PureIter,
)
from typing import (
    TypeVar,
)

_T = TypeVar("_T")


def to_tuple(piter: PureIter[_T], limit: int) -> FrozenList[_T]:
    n_items = 0
    items = []
    for i in piter:
        n_items += 1
        items.append(i)
        if n_items >= limit:
            break
    return tuple(items)


def assert_immutability(piter: PureIter[_T], only_count: bool = False) -> None:
    # for finite PureIter
    if only_count:
        assert sum(1 for _ in piter) == sum(1 for _ in piter)
    else:
        assert tuple(piter) == tuple(piter)


def assert_immutability_inf(piter: PureIter[_T]) -> None:
    # for infinite PureIter
    assert to_tuple(piter, 10) == to_tuple(piter, 10)



from __future__ import (
    annotations,
)

from dataclasses import (
    dataclass,
    field,
)
from fa_purity.frozen import (
    FrozenDict,
    FrozenList,
)
from fa_purity.json.primitive.core import (
    Primitive,
)
from fa_purity.union import (
    Coproduct,
)
from typing import (
    Callable,
    TypeVar,
    Union,
)

UnfoldedJVal = Union[
    FrozenDict[str, "JsonValue"], FrozenList["JsonValue"], Primitive
]
from fa_purity.json_2.primitive import (
    JsonPrimitive,
)

_T = TypeVar("_T")


@dataclass(frozen=True)
class _Private:
    pass


JsonObj = FrozenDict[str, "JsonValue"]


@dataclass(frozen=True)
class JsonValue:
    _private: _Private = field(repr=False, hash=False, compare=False)
    _value: Coproduct[JsonPrimitive, Coproduct[JsonObj, FrozenList[JsonValue]]]

    def map(
        self,
        primitive_case: Callable[[JsonPrimitive], _T],
        list_case: Callable[[FrozenList[JsonValue]], _T],
        dict_case: Callable[[FrozenDict[str, JsonValue]], _T],
    ) -> _T:
        return self._value.map(
            primitive_case, lambda c: c.map(dict_case, list_case)
        )

    @staticmethod
    def from_primitive(item: JsonPrimitive) -> JsonValue:
        return JsonValue(_Private(), Coproduct.inl(item))

    @staticmethod
    def from_json(item: JsonObj) -> JsonValue:
        return JsonValue(_Private(), Coproduct.inr(Coproduct.inl(item)))

    @staticmethod
    def from_list(item: FrozenList[JsonValue]) -> JsonValue:
        return JsonValue(_Private(), Coproduct.inr(Coproduct.inr(item)))



from dataclasses import (
    dataclass,
)
from typing import (
    Generic,
    TypeVar,
)

_T = TypeVar("_T")


@dataclass(frozen=True)
class Patch(Generic[_T]):
    # patch for https://github.com/python/mypy/issues/5485
    # upgrading mypy where the fix is included will deprecate this
    inner: _T

    @property
    def unwrap(self) -> _T:
        return self.inner



from __future__ import (
    annotations,
)

from dataclasses import (
    dataclass,
)
from deprecated import (
    deprecated,
)
from fa_purity.frozen import (
    FrozenDict,
    FrozenList,
)
from fa_purity.json.primitive.core import (
    Primitive,
)
from typing import (
    Callable,
    TypeVar,
    Union,
)

UnfoldedJVal = Union[
    FrozenDict[str, "JsonValue"], FrozenList["JsonValue"], Primitive
]
_T = TypeVar("_T")


@deprecated("Replaced. Use `fa_purity.json_2.value.JsonValue` instead")  # type: ignore[misc]
@dataclass(frozen=True)
class JsonValue:
    _value: UnfoldedJVal

    def unfold(
        self,
    ) -> UnfoldedJVal:
        return self._value

    def map(
        self,
        primitive_case: Callable[[Primitive], _T],
        list_case: Callable[[FrozenList[JsonValue]], _T],
        dict_case: Callable[[FrozenDict[str, JsonValue]], _T],
    ) -> _T:
        if isinstance(self._value, tuple):
            return list_case(self._value)
        if isinstance(self._value, FrozenDict):
            return dict_case(self._value)
        return primitive_case(self._value)



from fa_purity.pure_iter.factory import (
    from_flist,
    from_range,
    infinite_gen,
    infinite_range,
)
from tests.pure_iter._utils import (
    assert_immutability,
    assert_immutability_inf,
    to_tuple,
)


def test_flist() -> None:
    items = tuple(range(10))
    assert_immutability(from_flist(items))


def test_range() -> None:
    assert_immutability(from_range(range(10)))


def test_inf_range() -> None:
    assert_immutability_inf(infinite_range(3, 5))


def test_infinite_gen() -> None:
    items = infinite_gen(lambda x: x + 1, 0)
    assert_immutability_inf(items)
    assert to_tuple(items, 10) == tuple(range(10))



from ._inner import (
    InnerStream,
)
from fa_purity import (
    _iter_factory,
)
from fa_purity.cmd import (
    Cmd,
)
from fa_purity.pure_iter.core import (
    PureIter,
)
from fa_purity.stream.core import (
    Stream,
)
from typing import (
    Iterable,
    TypeVar,
)

_T = TypeVar("_T")


def unsafe_from_cmd(cmd: Cmd[Iterable[_T]]) -> Stream[_T]:
    # [WARNING] unsafe constructor
    # - Type-check cannot ensure its proper use
    # - Do not use until is strictly necessary
    # - Do unit test over the function defined by this
    #
    # As with `PureIter unsafe_from_cmd` the cmd must return a new iterable
    # object in each call to ensure that the stream is never consumed,
    # nevertheless they can be semanticly different iterables.
    return Stream(InnerStream(cmd))


def from_piter(piter: PureIter[Cmd[_T]]) -> Stream[_T]:
    items = InnerStream(_iter_factory.squash(piter))
    return Stream(items)



from fa_purity.result import (
    Result,
)
from fa_purity.union import (
    UnionFactory,
)
from typing import (
    Union,
)


def test_use_case_1() -> None:
    value = 245
    some: Result[int, str] = Result.success(value)
    result = (
        some.map(lambda i: i + 1)
        .bind(lambda i: Result.failure(f"fail {i}", int))
        .map(lambda i: i + 1)
        .alt(lambda x: f"{x} alt")
    )
    assert result.unwrap_fail() == f"fail {value + 1} alt"


def test_result_with_union() -> None:
    value = "hi"
    _union: UnionFactory[str, int] = UnionFactory()
    result: Result[Union[str, int], None] = Result.failure(None, int).map(
        _union.inr
    )
    result_2 = result.lash(
        lambda _: Result.success(value, type(None)).map(_union.inl)
    )
    assert result_2.unwrap() == value



from fa_purity.cmd import (
    Cmd,
)
from fa_purity.cmd.core import (
    CmdUnwrapper,
    new_cmd,
    unsafe_unwrap,
)
import pytest
from tempfile import (
    TemporaryFile,
)
from typing import (
    Callable,
    IO,
    NoReturn,
)


def _do_not_call() -> NoReturn:
    raise Exception("Cmd action should be only executed on compute phase")


def test_from_cmd() -> None:
    Cmd.from_cmd(_do_not_call)


def test_map() -> None:
    cmd = Cmd.from_cmd(lambda: 44).map(lambda i: i + 5)
    cmd.map(lambda _: _do_not_call())  # type: ignore
    Cmd.from_cmd(_do_not_call).map(lambda _: _)  # type: ignore

    def _verify(num: int) -> None:
        assert num == 49

    with pytest.raises(SystemExit):
        cmd.map(_verify).compute()


def test_bind() -> None:
    cmd = Cmd.from_cmd(lambda: 50)
    cmd2 = Cmd.from_cmd(lambda: 44).bind(lambda i: cmd.map(lambda x: x + i))
    cmd2.bind(lambda _: Cmd.from_cmd(_do_not_call))
    Cmd.from_cmd(_do_not_call).bind(lambda _: cmd)

    def _verify(num: int) -> None:
        assert num == 94

    with pytest.raises(SystemExit):
        cmd2.map(_verify).compute()


def test_apply() -> None:
    cmd = Cmd.from_cmd(lambda: 1)
    wrapped: Cmd[Callable[[int], int]] = Cmd.from_cmd(lambda: lambda x: x + 10)

    dead_end: Cmd[Callable[[int], NoReturn]] = Cmd.from_cmd(
        lambda: lambda _: _do_not_call()  # type: ignore
    )
    ex_falso_quodlibet: Callable[[NoReturn], int] = lambda _: 1
    wrap_no_return: Cmd[Callable[[NoReturn], int]] = Cmd.from_cmd(
        lambda: ex_falso_quodlibet
    )

    cmd.apply(dead_end)
    Cmd.from_cmd(_do_not_call).apply(wrap_no_return)

    def _verify(num: int) -> None:
        assert num == 11

    with pytest.raises(SystemExit):
        cmd.apply(wrapped).map(_verify).compute()


def _print_msg(msg: str, target: IO[str]) -> Cmd[None]:
    return Cmd.from_cmd(lambda: print(msg, file=target))


def test_use_case_1() -> None:
    with pytest.raises(SystemExit):
        with TemporaryFile("r+") as file:

            def _print(msg: str) -> Cmd[None]:
                return _print_msg(msg, file)

            in_val = Cmd.from_cmd(lambda: 245)
            some = in_val.map(lambda i: i + 1).map(str).bind(_print)
            _print("not called")
            pre = _print("Hello World!")
            try:
                pre.bind(lambda _: some).compute()
            except SystemExit as err:
                file.seek(0)
                assert file.readlines() == ["Hello World!\n", "246\n"]
                raise err


def test_new_cmd() -> None:
    state = {}

    def _mutate(val: int) -> None:
        state["temp"] = val

    change_1 = Cmd.from_cmd(lambda: _mutate(99)).map(lambda _: "1")
    change_2 = Cmd.from_cmd(lambda: _mutate(2)).map(lambda _: 2)

    def _action(unwrapper: CmdUnwrapper) -> int:
        x = unwrapper.act(change_2)
        y = unwrapper.act(change_1)
        return x + int(y)

    cmd1 = Cmd.new_cmd(_action)
    assert state.get("temp") is None
    assert unsafe_unwrap(cmd1) == 3
    assert state["temp"] == 99



# Iterable transforms
# should always return a new instance because Iterables are mutable
# result should be wrapped in a Cmd

from collections import (
    deque as deque_iter,
)
from fa_purity.cmd import (
    Cmd,
)
from fa_purity.cmd.core import (
    CmdUnwrapper,
    new_cmd,
)
from fa_purity.frozen import (
    freeze,
    FrozenList,
)
from itertools import (
    chain as _chain,
)
import more_itertools
from typing import (
    Callable,
    Iterable,
    Optional,
    Tuple,
    TypeVar,
)

_T = TypeVar("_T")
_R = TypeVar("_R")
_S = TypeVar("_S")


def chain(
    unchained: Iterable[Iterable[_T]],
) -> Iterable[_T]:
    return _chain.from_iterable(unchained)


def chunked(items: Iterable[_T], size: int) -> Iterable[FrozenList[_T]]:
    return map(freeze, more_itertools.chunked(items, size))


def deque(items: Iterable[_T]) -> None:
    deque_iter(items, maxlen=0)


def filter_none(items: Iterable[Optional[_T]]) -> Iterable[_T]:
    return (i for i in items if i is not None)


def find_first(
    criteria: Callable[[_T], bool], items: Iterable[_T]
) -> Optional[_T]:
    for item in items:
        if criteria(item):
            return item
    return None


def squash(items: Iterable[Cmd[_T]]) -> Cmd[Iterable[_T]]:
    def _action(unwrapper: CmdUnwrapper) -> Iterable[_T]:
        for item in items:
            yield unwrapper.act(item)

    return Cmd.new_cmd(_action)


def until_none(items: Iterable[Optional[_T]]) -> Iterable[_T]:
    for item in items:
        if item is None:
            break
        yield item


def infinite_gen(function: Callable[[_T], _T], init: _T) -> Iterable[_T]:
    yield init
    current = init
    while True:
        current = function(current)
        yield current


def gen_next(
    function: Callable[[_S, _T], Tuple[_S, _R]],
    init_state: _S,
    iter: Iterable[_T],
) -> Iterable[_R]:
    state = init_state
    for i in iter:
        state, current = function(state, i)
        yield current



from ._core import (
    JsonPrimitive,
)
from ._factory import (
    JsonPrimitiveFactory,
    Primitive,
)
from ._transform import (
    JsonPrimitiveUnfolder,
)

__all__ = [
    "Primitive",
    "JsonPrimitive",
    "JsonPrimitiveFactory",
    "JsonPrimitiveUnfolder",
]



from dataclasses import (
    dataclass,
)
from typing import (
    NoReturn,
)


@dataclass(frozen=True)
class LibraryBug(Exception):
    traceback: Exception

    def __str__(self) -> str:
        return f"If raised then there is a bug in the `fa_purity` library"


def raise_exception(err: Exception) -> NoReturn:
    raise err



from __future__ import (
    annotations,
)

from dataclasses import (
    dataclass,
)
from fa_purity import (
    _iter_factory,
)
from fa_purity.cmd import (
    unsafe_unwrap,
)
from fa_purity.cmd.core import (
    Cmd,
)
from fa_purity.frozen import (
    FrozenList,
)
from fa_purity.maybe import (
    Maybe,
)
from fa_purity.pure_iter._inner import (
    InnerPureIter,
)
import functools
from typing import (
    Callable,
    Generic,
    Iterator,
    Tuple,
    TypeVar,
)

_T = TypeVar("_T")
_R = TypeVar("_R")
_S = TypeVar("_S")


@dataclass(frozen=True)
class PureIter(Generic[_T]):
    # unsafe_unwrap use is safe due to iters equivalence
    _inner: InnerPureIter[_T]

    def generate(
        self, function: Callable[[_S, _T], Tuple[_S, _R]], init_state: _S
    ) -> PureIter[_R]:
        """
        Apply the function over each item.
        The argument state (_S) comes from a previous computation of the function over
        a previous state-item pair or from `init_state` if is the first element.
        """
        draft: InnerPureIter[_R] = InnerPureIter(
            self._inner.new_iter.map(
                lambda i: _iter_factory.gen_next(function, init_state, i)
            )
        )
        return PureIter(draft)

    def enumerate(self, init: int) -> PureIter[Tuple[int, _T]]:
        return self.generate(lambda i, t: (i + 1, (i, t)), init)

    def map(self, function: Callable[[_T], _R]) -> PureIter[_R]:
        return self.generate(lambda _, t: (None, function(t)), None)

    def reduce(self, function: Callable[[_R, _T], _R], init: _R) -> _R:
        return unsafe_unwrap(
            self._inner.new_iter.map(
                lambda i: functools.reduce(function, i, init)
            )
        )

    def bind(self, function: Callable[[_T], PureIter[_R]]) -> PureIter[_R]:
        unchained = self.map(function)
        draft: InnerPureIter[_R] = InnerPureIter(
            Cmd.from_cmd(lambda: _iter_factory.chain(unchained))
        )
        return PureIter(draft)

    def filter(self, function: Callable[[_T], bool]) -> PureIter[_T]:
        draft = InnerPureIter(
            self._inner.new_iter.map(lambda i: iter(filter(function, i)))
        )
        return PureIter(draft)

    def find_first(self, criteria: Callable[[_T], bool]) -> Maybe[_T]:
        result = self._inner.new_iter.map(
            lambda i: _iter_factory.find_first(criteria, i)
        ).map(lambda x: Maybe.from_optional(x))
        return unsafe_unwrap(result)

    def chunked(self, size: int) -> PureIter[FrozenList[_T]]:
        draft = InnerPureIter(
            self._inner.new_iter.map(lambda i: _iter_factory.chunked(i, size))
        )
        return PureIter(draft)

    def to_list(self) -> FrozenList[_T]:
        return tuple(self)

    def transform(self, function: Callable[[PureIter[_T]], _R]) -> _R:
        return function(self)

    def __iter__(self) -> Iterator[_T]:
        return iter(unsafe_unwrap(self._inner.new_iter))



from fa_purity.cmd import (
    Cmd,
)
from fa_purity.frozen import (
    FrozenList,
)
from fa_purity.pure_iter.factory import (
    from_flist,
    from_range,
)
from fa_purity.stream.core import (
    Stream,
)
from fa_purity.stream.factory import (
    from_piter,
)
from fa_purity.stream.transform import (
    chain,
    consume,
    filter_opt,
    squash,
    until_none,
)
import pytest
from tests.stream._utils import (
    assert_different_iter,
    rand_int,
)
from typing import (
    Optional,
)


def _equal(a: int, b: int) -> None:
    assert a == b


def _mock_stream_opt(
    size: int, none_at: Optional[int]
) -> Stream[Optional[int]]:
    items = from_range(range(size)).map(
        lambda i: rand_int().map(lambda r: r if none_at != i else None)
    )
    return from_piter(items)


def test_chain_1() -> None:
    base = from_flist((1, 2, 3))
    items = base.map(lambda i: Cmd.from_cmd(lambda: i))
    stm = from_piter(items)
    unchained = stm.map(lambda _: base)
    chained = chain(unchained)
    assert_different_iter(chained)

    def _verify(elements: FrozenList[int]) -> None:
        assert elements == (1, 2, 3) * 3

    with pytest.raises(SystemExit):
        chained.to_list().map(_verify).compute()


def test_chain_2() -> None:
    base = from_flist((1, 2, 3))
    items = base.map(lambda i: Cmd.from_cmd(lambda: i))
    stm = from_piter(items)
    unchained = base.map(lambda _: stm)
    chained = chain(unchained)
    assert_different_iter(chained)

    def _verify(elements: FrozenList[int]) -> None:
        assert elements == (1, 2, 3) * 3

    with pytest.raises(SystemExit):
        chained.to_list().map(_verify).compute()


def test_consume() -> None:
    mutable = [0]

    def _mutate(i: int) -> Cmd[None]:
        def _action() -> None:
            mutable[0] = i

        return Cmd.from_cmd(_action)

    items = from_range(range(10)).map(lambda i: Cmd.from_cmd(lambda: i))
    stm = from_piter(items).map(lambda i: _mutate(i))
    result = consume(stm)

    def _verify() -> None:
        assert mutable[0] == 9

    assert mutable[0] == 0
    with pytest.raises(SystemExit):
        result.map(lambda _: _verify()).compute()


def test_squash() -> None:
    items = from_range(range(10)).map(lambda _: rand_int())
    stm = from_piter(items).map(lambda _: rand_int())
    result = squash(stm)
    assert_different_iter(result)


def test_filter_opt() -> None:
    stm = _mock_stream_opt(10, 5)
    stm_len = filter_opt(stm).to_list().map(len)
    with pytest.raises(SystemExit):
        stm_len.map(lambda n: _equal(n, 9)).compute()


def test_until_none() -> None:
    stm = _mock_stream_opt(10, 5)
    stm_len = until_none(stm).to_list().map(len)
    with pytest.raises(SystemExit):
        stm_len.map(lambda n: _equal(n, 5)).compute()



from fa_purity.frozen import (
    FrozenList,
)
from fa_purity.result.core import (
    Result,
    UnwrapError,
)
from typing import (
    cast,
    TypeVar,
)

_S = TypeVar("_S")
_F = TypeVar("_F")


def all_ok(items: FrozenList[Result[_S, _F]]) -> Result[FrozenList[_S], _F]:
    try:
        return Result.success(tuple(i.unwrap() for i in items))
    except UnwrapError as err:
        _err = cast(UnwrapError[_S, _F], err)
        return Result.failure(_err.container.unwrap_fail())



from ._core import (
    JsonObj,
    JsonValue,
)
from dataclasses import (
    dataclass,
)
from fa_purity.frozen import (
    FrozenDict,
    FrozenList,
)
from fa_purity.json_2.primitive import (
    JsonPrimitiveFactory,
    JsonPrimitiveUnfolder,
    Primitive,
)
from fa_purity.result import (
    Result,
)
from fa_purity.result.core import (
    ResultE,
)
from fa_purity.utils import (
    raise_exception,
)
import simplejson
from typing import (
    Any,
    cast,
    Dict,
    IO,
    List,
    Optional,
    TypeVar,
    Union,
)

_T = TypeVar("_T")


class _HandledException(Exception):
    pass


@dataclass(frozen=True)
class UnfoldedFactory:
    @staticmethod
    def from_list(
        raw: Union[List[Primitive], FrozenList[Primitive]]
    ) -> FrozenList[JsonValue]:
        return tuple(
            JsonValue.from_primitive(JsonPrimitiveFactory.from_raw(item))
            for item in raw
        )

    @staticmethod
    def from_dict(
        raw: Union[Dict[str, Primitive], FrozenDict[str, Primitive]]
    ) -> JsonObj:
        return FrozenDict(
            {
                key: JsonValue.from_primitive(
                    JsonPrimitiveFactory.from_raw(val)
                )
                for key, val in raw.items()
            }
        )


@dataclass(frozen=True)
class JsonValueFactory:
    @staticmethod
    def from_list(
        raw: Union[List[Primitive], FrozenList[Primitive]]
    ) -> JsonValue:
        return JsonValue.from_list(UnfoldedFactory.from_list(raw))

    @staticmethod
    def from_dict(
        raw: Union[Dict[str, Primitive], FrozenDict[str, Primitive]]
    ) -> JsonValue:
        return JsonValue.from_json(UnfoldedFactory.from_dict(raw))

    @classmethod
    def from_any(cls, raw: Optional[_T]) -> ResultE[JsonValue]:
        if isinstance(raw, (FrozenDict, dict)):
            try:
                json_dict = FrozenDict(
                    {
                        JsonPrimitiveFactory.from_any(key)
                        .bind(JsonPrimitiveUnfolder.to_str)
                        .alt(_HandledException)
                        .alt(raise_exception)
                        .unwrap(): cls.from_any(val)
                        .alt(_HandledException)
                        .alt(raise_exception)
                        .unwrap()
                        for key, val in raw.items()
                    }
                )
                return Result.success(JsonValue.from_json(json_dict))
            except _HandledException as err:
                return Result.failure(Exception(err))
        if isinstance(raw, (list, tuple)):
            try:
                json_list = tuple(
                    cls.from_any(item)
                    .alt(_HandledException)
                    .alt(raise_exception)
                    .unwrap()
                    for item in raw
                )
                return Result.success(JsonValue.from_list(json_list))
            except _HandledException as err:
                return Result.failure(Exception(err))
        return JsonPrimitiveFactory.from_any(raw).map(JsonValue.from_primitive)

    @classmethod
    def _from_raw_dict(cls, raw: Dict[str, Any]) -> ResultE[JsonObj]:  # type: ignore[misc]
        err = Result.failure(Exception(TypeError("Not a `JsonObj`")), JsonObj)
        return cls.from_any(raw).bind(  # type: ignore[misc]
            lambda jv: jv.map(
                lambda _: err,
                lambda _: err,
                lambda x: Result.success(x),
            )
        )

    @classmethod
    def loads(cls, raw: str) -> ResultE[JsonObj]:
        raw_json = cast(Dict[str, Any], simplejson.loads(raw))  # type: ignore[misc]
        return cls._from_raw_dict(raw_json)  # type: ignore[misc]

    @classmethod
    def load(cls, raw: IO[str]) -> ResultE[JsonObj]:
        raw_json = cast(Dict[str, Any], simplejson.load(raw))  # type: ignore[misc]
        return cls._from_raw_dict(raw_json)  # type: ignore[misc]



from __future__ import (
    annotations,
)

from dataclasses import (
    dataclass,
    field,
)
from datetime import (
    datetime,
    timedelta,
    timezone,
)
from fa_purity import (
    ResultE,
)
from fa_purity.cmd.core import (
    Cmd,
)
from fa_purity.result import (
    Result,
)


@dataclass(frozen=True)
class _Private:
    pass


@dataclass(frozen=True)
class DatetimeTZ:
    _private: _Private = field(repr=False, hash=False, compare=False)
    date_time: datetime

    @staticmethod
    def assert_tz(time: datetime) -> ResultE[DatetimeTZ]:
        if time.tzinfo is not None:
            return Result.success(DatetimeTZ(_Private(), time))
        err = ValueError("datetime must have a timezone")
        return Result.failure(err, DatetimeTZ).alt(Exception)

    def __add__(self, delta: timedelta) -> DatetimeTZ:
        return self.assert_tz(self.date_time + delta).unwrap()

    def __sub__(self, delta: timedelta) -> DatetimeTZ:
        return self.assert_tz(self.date_time - delta).unwrap()


@dataclass(frozen=True)
class DatetimeUTC:
    _private: _Private = field(repr=False, hash=False, compare=False)
    date_time: datetime

    @staticmethod
    def assert_utc(time: datetime | DatetimeTZ) -> ResultE[DatetimeUTC]:
        _time = time if isinstance(time, datetime) else time.date_time
        if _time.tzinfo == timezone.utc:
            return Result.success(DatetimeUTC(_Private(), _time))
        err = ValueError(
            f"datetime must have UTC timezone but got {_time.tzinfo}"
        )
        return Result.failure(err, DatetimeUTC).alt(Exception)

    def __add__(self, delta: timedelta) -> DatetimeUTC:
        return self.assert_utc(self.date_time + delta).unwrap()

    def __sub__(self, delta: timedelta) -> DatetimeUTC:
        return self.assert_utc(self.date_time - delta).unwrap()


@dataclass(frozen=True)
class DatetimeFactory:
    EPOCH_START: DatetimeUTC = DatetimeUTC.assert_utc(
        datetime.fromtimestamp(0, timezone.utc)
    ).unwrap()

    @staticmethod
    def new_utc(
        year: int,
        month: int,
        day: int,
        hour: int,
        minute: int,
        second: int,
        microsecond: int,
    ) -> DatetimeUTC:
        return DatetimeUTC.assert_utc(
            datetime(
                year,
                month,
                day,
                hour,
                minute,
                second,
                microsecond,
                tzinfo=timezone.utc,
            )
        ).unwrap()

    @staticmethod
    def new_tz(
        year: int,
        month: int,
        day: int,
        hour: int,
        minute: int,
        second: int,
        microsecond: int,
        zone: timezone,
    ) -> DatetimeTZ:
        return DatetimeTZ.assert_tz(
            datetime(
                year,
                month,
                day,
                hour,
                minute,
                second,
                microsecond,
                tzinfo=zone,
            )
        ).unwrap()

    @staticmethod
    def to_tz(
        date_time: datetime | DatetimeUTC, time_zone: timezone
    ) -> DatetimeTZ:
        item = (
            date_time
            if isinstance(date_time, datetime)
            else date_time.date_time
        )
        return DatetimeTZ.assert_tz(item.astimezone(time_zone)).unwrap()

    @classmethod
    def to_utc(cls, date_time: DatetimeTZ) -> DatetimeUTC:
        return DatetimeUTC.assert_utc(
            cls.to_tz(date_time.date_time, timezone.utc)
        ).unwrap()

    @staticmethod
    def date_now() -> Cmd[DatetimeUTC]:
        return Cmd.from_cmd(
            lambda: DatetimeUTC.assert_utc(datetime.now(timezone.utc)).unwrap()
        )



from deprecated import (
    deprecated,
)
from fa_purity.frozen import (
    FrozenDict,
    unfreeze,
)
from fa_purity.json.factory import (
    JsonObj,
)
from fa_purity.json.value import (
    transform as jval_transform,
)
from fa_purity.json.value.core import (
    JsonValue,
)
from simplejson import (
    dumps as _dumps,
    JSONEncoder,
)
from typing import (
    Any,
    Dict,
    TypeVar,
)


@deprecated("Replaced. Use `fa_purity.json_2.Unfolder` instead")  # type: ignore[misc]
def to_raw(json_obj: JsonObj) -> Dict[str, Any]:
    return {key: jval_transform.to_raw(val) for key, val in json_obj.items()}  # type: ignore[misc]


_T = TypeVar("_T")


class _JsonEncoder(JSONEncoder):
    def default(self: JSONEncoder, o: _T) -> Any:  # type: ignore[misc]
        if isinstance(o, FrozenDict):
            return unfreeze(o)  # type: ignore[misc]
        if isinstance(o, JsonValue):
            return o.unfold()
        return JSONEncoder.default(self, o)  # type: ignore[misc]


@deprecated("Replaced. Use `fa_purity.json_2.Unfolder` instead")  # type: ignore[misc]
def dumps(obj: JsonObj) -> str:
    return _dumps(obj, cls=_JsonEncoder)  # type: ignore[misc]



from __future__ import (
    annotations,
)

from dataclasses import (
    dataclass,
)
from deprecated import (
    deprecated,
)
from fa_purity.frozen import (
    freeze,
    FrozenDict,
    FrozenList,
)
from fa_purity.json.errors import (
    invalid_type,
)
from fa_purity.json.errors.invalid_type import (
    InvalidType,
)
from fa_purity.json.primitive.core import (
    is_primitive,
    NotNonePrimTvar,
    Primitive,
)
from fa_purity.json.primitive.factory import (
    to_primitive as _to_primitive,
)
from fa_purity.json.value.core import (
    JsonValue,
    UnfoldedJVal,
)
from fa_purity.result import (
    Result,
)
from fa_purity.result.core import (
    ResultE,
)
from fa_purity.result.factory import (
    ResultFactory,
    try_get,
)
from fa_purity.result.transform import (
    all_ok,
)
from fa_purity.union import (
    UnionFactory,
)
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
)

_T = TypeVar("_T")
UnfoldResult = Result[_T, InvalidType]  # type: ignore[misc]


@deprecated("Replaced. Use `fa_purity.json_2.value.Unfolder` instead")  # type: ignore[misc]
@dataclass(frozen=True)
class Unfolder:
    jval: JsonValue

    @property
    def value(self) -> UnfoldedJVal:
        return self.jval.unfold()

    def get(self, key: str) -> ResultE[JsonValue]:
        return self.to_json().alt(Exception).bind(lambda d: try_get(d, key))

    def uget(self, key: str) -> ResultE[Unfolder]:
        return self.get(key).map(lambda j: Unfolder(j))

    def to_primitive(
        self, prim_type: Type[NotNonePrimTvar]
    ) -> Result[NotNonePrimTvar, InvalidType]:
        return _to_primitive(self.value, prim_type)  # type: ignore[arg-type]

    def to_any_primitive(self) -> ResultE[Primitive]:
        factory: ResultFactory[Primitive, Exception] = ResultFactory()
        fail = factory.failure(
            invalid_type.new("to_any_primitive", "Primitive", self.value)
        )
        return self.jval.map(
            factory.success,
            lambda _: fail,
            lambda _: fail,
        )

    def to_none(self) -> Result[None, InvalidType]:
        if self.value is None:
            return Result.success(self.value)
        return Result.failure(invalid_type.new("to_none", "None", self.value))

    def to_list(self) -> UnfoldResult[FrozenList[JsonValue]]:
        factory: ResultFactory[
            FrozenList[JsonValue], InvalidType
        ] = ResultFactory()
        fail = factory.failure(
            invalid_type.new("to_list", "FrozenList[JsonValue]", self.value)
        )
        return self.jval.map(
            lambda _: fail,
            factory.success,
            lambda _: fail,
        )

    def to_optional(
        self, transform: Callable[[Unfolder], UnfoldResult[_T]]
    ) -> UnfoldResult[Optional[_T]]:
        _union: UnionFactory[_T, None] = UnionFactory()
        return (
            self.to_none()
            .map(_union.inr)
            .lash(lambda _: transform(self).map(_union.inl))
        )

    def to_opt_list(self) -> UnfoldResult[Optional[FrozenList[JsonValue]]]:
        return self.to_optional(lambda j: j.to_list())

    def to_list_of(
        self, prim_type: Type[NotNonePrimTvar]
    ) -> UnfoldResult[FrozenList[NotNonePrimTvar]]:
        return (
            self.to_list()
            .map(
                lambda l: tuple(
                    _to_primitive(i.unfold(), prim_type) for i in l  # type: ignore[arg-type]
                )
            )
            .bind(all_ok)
        )

    def to_unfolder_list(self) -> UnfoldResult[FrozenList[Unfolder]]:
        return self.to_list().map(lambda l: tuple(Unfolder(i) for i in l))

    def to_json(self) -> UnfoldResult[FrozenDict[str, JsonValue]]:
        factory: ResultFactory[
            FrozenDict[str, JsonValue], InvalidType
        ] = ResultFactory()
        fail = factory.failure(
            invalid_type.new(
                "to_json", "FrozenDict[str, JsonValue]", self.value
            )
        )
        return self.jval.map(
            lambda _: fail,
            lambda _: fail,
            factory.success,
        )

    def to_unfolder_dict(self) -> UnfoldResult[FrozenDict[str, Unfolder]]:
        return (
            self.to_json()
            .map(lambda d: {k: Unfolder(v) for k, v in d.items()})
            .map(freeze)
        )

    def to_dict_of(
        self, prim_type: Type[NotNonePrimTvar]
    ) -> UnfoldResult[FrozenDict[str, NotNonePrimTvar]]:
        return (
            self.to_json()
            .map(
                lambda d: tuple(
                    _to_primitive(val.unfold(), prim_type).map(  # type: ignore[arg-type]
                        lambda v: (key, v)
                    )
                    for key, val in d.items()
                )
            )
            .bind(lambda x: all_ok(x))
            .map(lambda x: FrozenDict(dict(x)))
        )


@deprecated("Replaced. Use `fa_purity.json_2.value.Unfolder` instead")  # type: ignore[misc]
def to_raw(value: JsonValue) -> Union[Dict[str, Any], List[Any], Primitive]:
    return value.map(
        lambda x: x,
        lambda items: [to_raw(i) for i in items],  # type: ignore[misc]
        lambda dict_obj: {key: to_raw(val) for key, val in dict_obj.items()},  # type: ignore[misc]
    )



from fa_purity.cmd.core import (
    Cmd,
    unsafe_unwrap,
)
from fa_purity.frozen import (
    FrozenList,
)
from typing import (
    Callable,
    TypeVar,
)

_T = TypeVar("_T")
_R = TypeVar("_R")


Mapper = Callable[[Callable[[_T], _R], FrozenList[_T]], FrozenList[_R]]


def merge(
    mapper: Mapper[Cmd[_T], _T], items: FrozenList[Cmd[_T]]
) -> Cmd[FrozenList[_T]]:
    def _action() -> FrozenList[_T]:
        return mapper(lambda i: unsafe_unwrap(i), items)

    return Cmd.from_cmd(_action)


def serial_merge(items: FrozenList[Cmd[_T]]) -> Cmd[FrozenList[_T]]:
    return merge(lambda f, i: tuple(map(f, i)), items)



from deprecated import (
    deprecated,
)
from fa_purity.json.errors import (
    invalid_type,
)
from fa_purity.json.errors.invalid_type import (
    InvalidType,
)
from fa_purity.json.primitive.core import (
    NotNonePrimTvar,
)
from fa_purity.result import (
    Result,
)
from typing import (
    Optional,
    Type,
    TypeVar,
)

_T = TypeVar("_T")


@deprecated("Replaced. Use `fa_purity.json_2.primitive.JsonPrimitiveUnfolder` instead")  # type: ignore[misc]
def to_primitive(
    raw: _T, prim_type: Type[NotNonePrimTvar]
) -> Result[NotNonePrimTvar, InvalidType]:
    if isinstance(raw, prim_type):
        return Result.success(raw)
    return Result.failure(
        invalid_type.new("to_primitive", str(prim_type), raw)
    )


@deprecated("Replaced. Use `fa_purity.json_2.primitive.JsonPrimitiveUnfolder` instead")  # type: ignore[misc]
def to_opt_primitive(
    raw: _T, prim_type: Type[NotNonePrimTvar]
) -> Result[Optional[NotNonePrimTvar], InvalidType]:
    if raw is None or isinstance(raw, prim_type):
        return Result.success(raw)
    return Result.failure(
        invalid_type.new("to_opt_primitive", f"{prim_type} | None", raw)
    )



from ._core import (
    JsonPrimitive,
)
from dataclasses import (
    dataclass,
)
from decimal import (
    Decimal,
)
from fa_purity.result import (
    Result,
)
from fa_purity.result.core import (
    ResultE,
)
from typing import (
    Optional,
    TypeVar,
    Union,
)

_T = TypeVar("_T")
Primitive = Union[str, int, float, Decimal, bool, None]


@dataclass(frozen=True)
class JsonPrimitiveFactory:
    @staticmethod
    def from_any(
        raw: Optional[_T],
    ) -> ResultE[JsonPrimitive]:
        if raw is None:
            return Result.success(JsonPrimitive.empty())
        if isinstance(raw, bool):
            return Result.success(JsonPrimitive.from_bool(raw))
        if isinstance(raw, str):
            return Result.success(JsonPrimitive.from_str(raw))
        if isinstance(raw, int):
            return Result.success(JsonPrimitive.from_int(raw))
        if isinstance(raw, float):
            return Result.success(JsonPrimitive.from_float(raw))
        if isinstance(raw, Decimal):
            return Result.success(JsonPrimitive.from_decimal(raw))
        return Result.failure(
            Exception(TypeError("Cannot convert to `JsonPrimitive`"))
        )

    @staticmethod
    def from_raw(raw: Primitive) -> JsonPrimitive:
        if raw is None:
            return JsonPrimitive.empty()
        if isinstance(raw, bool):
            return JsonPrimitive.from_bool(raw)
        if isinstance(raw, str):
            return JsonPrimitive.from_str(raw)
        if isinstance(raw, int):
            return JsonPrimitive.from_int(raw)
        if isinstance(raw, float):
            return JsonPrimitive.from_float(raw)
        if isinstance(raw, Decimal):
            return JsonPrimitive.from_decimal(raw)



from ._core import (
    JsonObj,
    JsonValue,
)
from dataclasses import (
    dataclass,
)
from fa_purity.frozen import (
    FrozenDict,
    FrozenList,
    unfreeze,
)
from fa_purity.json_2.primitive import (
    JsonPrimitive,
    JsonPrimitiveUnfolder,
    Primitive,
)
from fa_purity.result import (
    Result,
)
from fa_purity.result.core import (
    ResultE,
)
from fa_purity.result.factory import (
    try_get,
)
from fa_purity.result.transform import (
    all_ok,
)
from fa_purity.union import (
    UnionFactory,
)
from simplejson import (
    dumps as _dumps,
    JSONEncoder,
)
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    TypeVar,
    Union,
)

_T = TypeVar("_T")


class _JsonEncoder(JSONEncoder):
    def default(self: JSONEncoder, o: _T) -> Any:  # type: ignore[misc]
        if isinstance(o, FrozenDict):
            return unfreeze(o)  # type: ignore[misc]
        if isinstance(o, JsonValue):
            result = o.map(
                lambda x: x.map(
                    lambda y: y,
                    lambda y: y,
                    lambda y: y,
                    lambda y: y,
                    lambda y: y,
                    lambda: None,
                ),
                lambda x: x,
                lambda x: x,
            )
            return result
        return JSONEncoder.default(self, o)  # type: ignore[misc]


@dataclass(frozen=True)
class Unfolder:
    @staticmethod
    def to_primitive(item: JsonValue) -> ResultE[JsonPrimitive]:
        fail: ResultE[JsonPrimitive] = Result.failure(
            Exception(
                TypeError("Expected `JsonPrimitive` in unfolded `JsonValue`")
            )
        )
        return item.map(
            lambda x: Result.success(x),
            lambda _: fail,
            lambda _: fail,
        )

    @staticmethod
    def to_list(item: JsonValue) -> ResultE[FrozenList[JsonValue]]:
        fail: ResultE[FrozenList[JsonValue]] = Result.failure(
            Exception(
                TypeError(
                    "Expected `FrozenList[JsonValue]` in unfolded `JsonValue`"
                )
            )
        )
        return item.map(
            lambda _: fail,
            lambda x: Result.success(x),
            lambda _: fail,
        )

    @staticmethod
    def to_json(item: JsonValue) -> ResultE[JsonObj]:
        fail: ResultE[JsonObj] = Result.failure(
            Exception(TypeError("Expected `JsonObj` in unfolded `JsonValue`"))
        )
        return item.map(
            lambda _: fail,
            lambda _: fail,
            lambda x: Result.success(x),
        )

    @staticmethod
    def transform_list(
        items: FrozenList[JsonValue],
        transform: Callable[[JsonValue], ResultE[_T]],
    ) -> ResultE[FrozenList[_T]]:
        return all_ok(tuple(transform(i) for i in items))

    @staticmethod
    def transform_json(
        item: JsonObj, transform: Callable[[JsonValue], ResultE[_T]]
    ) -> ResultE[FrozenDict[str, _T]]:
        key_values = tuple(
            transform(val).map(lambda p: (key, p)) for key, val in item.items()
        )
        return all_ok(key_values).map(lambda x: FrozenDict(dict(x)))

    @staticmethod
    def dumps(obj: JsonObj) -> str:
        return _dumps(obj, cls=_JsonEncoder)  # type: ignore[misc]

    @classmethod
    def get(cls, item: JsonValue, key: str) -> ResultE[JsonValue]:
        return cls.to_json(item).alt(Exception).bind(lambda d: try_get(d, key))

    @classmethod
    def to_list_of(
        cls, item: JsonValue, transform: Callable[[JsonValue], ResultE[_T]]
    ) -> ResultE[FrozenList[_T]]:
        return cls.to_list(item).bind(
            lambda i: cls.transform_list(i, transform)
        )

    @classmethod
    def to_dict_of(
        cls, item: JsonValue, transform: Callable[[JsonValue], ResultE[_T]]
    ) -> ResultE[FrozenDict[str, _T]]:
        return cls.to_json(item).bind(
            lambda i: cls.transform_json(i, transform)
        )

    @classmethod
    def to_optional(
        cls, item: JsonValue, transform: Callable[[JsonValue], ResultE[_T]]
    ) -> ResultE[Optional[_T]]:
        _union: UnionFactory[_T, None] = UnionFactory()
        return (
            cls.to_primitive(item)
            .bind(JsonPrimitiveUnfolder.to_none)
            .map(_union.inr)
            .lash(lambda _: transform(item).map(_union.inl))
        )

    @classmethod
    def to_raw(cls, value: JsonValue) -> Union[Dict[str, Any], List[Any], Primitive]:  # type: ignore[misc]
        def _cast(item: Primitive) -> Primitive:
            # cast used for helping mypy to infer the correct return type
            return item

        return value.map(
            lambda p: p.map(
                lambda x: _cast(x),
                lambda x: _cast(x),
                lambda x: _cast(x),
                lambda x: _cast(x),
                lambda x: _cast(x),
                lambda: _cast(None),
            ),
            lambda items: [cls.to_raw(i) for i in items],  # type: ignore[misc]
            lambda dict_obj: {key: cls.to_raw(val) for key, val in dict_obj.items()},  # type: ignore[misc]
        )
